<?php

// DBRow.class.php - PHPOF database row class
//
// $Id: DBRow.class.php 2264 2018-06-01 19:49:11Z tim $
//
// (c)2001-2018 Tim Jackson (tim@timj.co.uk)
//
// This library is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public
// License as published by the Free Software Foundation; either
// version 2.1 of the License, or (at your option) any later version.
//
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// Lesser General Public License for more details.
// 
// You should have received a copy of the GNU Lesser General Public
// License along with this library; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA


// --------------------------------------------------------------------------------------------
// INFO:
// This is the way binary data works:
// If the column is EMPTY, NULL or ZERO there is no binary data
// *Any* (internally-generated?) 'SELECT' query will NOT place the actual binary data in an object; it will
// replace it with the placeholder PHPOF_BINARYDATA_PRESENT. To retrieve the data, it will be necessary
// to call DBRow::getBinaryData().
//
// When inserting binary data, you need to FIRST have the row which will contain the data
// already inserted in the database. Then you need to call DBRow::storeBinaryData() or DBRow::handleUpload() which 
// will insert the data directly into the database and fill the object attribute for the
// relevant column with PHPOF_BINARYDATA_PRESENT.
//
// Ditto for updating binary data in a row.
//
// If you want to remove the binary data from a row, mark the column with PHPOF_BINARYDATA_DELETE and
// next time you do a DBRow::update() the data will be deleted.

// Used for binary data
define('DBROW_GET',0);
define('DBROW_CHECKONLY',1);

// For DBRow::clearFields();
define('DBROW_SET_TO_NULL',0);
define('DBROW_UNSET',1);
define('DBROW_KEY_SAVE',-1);
define('DBROW_KEY_CLEAR',-2);

// For DBRow::move()
define('DBROW_MOVE_UP','up');
define('DBROW_MOVE_DOWN','down');
define('DBROW_SORT_TOP',-1);
define('DBROW_SORT_BOTTOM',-2);

define('DBROW_KEY_NONE','!phpof-nokey!');

require_once('DBTable.class.php');

// General TODO: add support for composite keys

/**
 * This class is used for modelling rows within a database table.
 *
 */
class DBRow {	// friend DB_result

	var $_table;		// Contains table object which this row belongs to
	var $_audit;		// Array containing the original values of columns

	/**
	 * Set to true to skip audit recording for this row. Used whilst inserting 
	 * audit entries so we don't get infinite recursion (auditing the audit etc.)
	 *
	 * @var bool
	 */
	var $_skip_audit = false;
	var $_audit_notes;	// 
	var $_audit_owner;
	var $_Hook_FetchIntoObject_pre;			// Do NOT use '$this->'
	var $_Hook_FetchIntoObject_post; 		// Do NOT use '$this->'
	var $_Hook_Constructor_pre;
	var $_Hook_Constructor_post;
	var $_Hook_GetDetails_pre;
	var $_Hook_GetDetails_post;
	var $_Hook_Store_pre;
	var $_Hook_Store_post;
	var $_Hook_Insert_pre;
	var $_Hook_Insert_post;
	var $_Hook_Update_pre;
	var $_Hook_Update_post;
	var $_Hook_Delete_pre;
	var $_Hook_Delete_post;
	
	// DescriptiveNameExpression should contain an expression which will provide a human-readable
	// description of the row when evaluated. It will be evaluated in the context of the DBRow object, so 
	// can contain references to "$this"
	var $DescriptiveNameExpression = '';
	
	// DescriptiveName will be filled when ::getDescriptiveName() is called
	var $DescriptiveName;
	
	// This is set to TRUE on a sucessful getDetails();
	var $RetrievedOK = false;
	
	/**
	* Initialising constructor.
	* 
	* @param DBTable $table  The table which this row belongs to
	* @param string $key  Primary key of row to fetch (using DBRow::getDetails()) into the object upon instantiation
	* @return void
	* @access public
	*/
	function DBRow (&$table, $key = DBROW_KEY_NONE)
	{
		global $PHPOF_INITIALISE_PROPERTIES_TO_NULL;
		
		eval($this->_Hook_Constructor_pre);
		
		// TODO? If foreign relations are defined, add a _Hook_Delete_pre, which cascades the delete
		
		if (!is_object($table)) trigger_error('DBRow constructor was passed a non-object instead of DBTable object',E_USER_ERROR);
		
		$this->_table = &$table;
		
		if ($PHPOF_INITIALISE_PROPERTIES_TO_NULL) $this->clearFields(DBROW_SET_TO_NULL, DBROW_KEY_CLEAR);
		
		// For backwards compat with anything that might try to pass in null or empty string as a key to mean no key
		if ($key === null || $key === '') {
			$key = DBROW_KEY_NONE;
		}
		
		settype($key,'string');	// This is necessary to make sure that numeric keys match below expression
		if ($key != DBROW_KEY_NONE) {
			$this->{$this->_table->_key} = $key;
			$this->getDetails();
		}
		
		// Default to using the 'name' column as the descriptive name, if it exists
		if (empty($this->DescriptiveNameExpression) && in_array('name',$this->_table->_fields)) {
			$this->DescriptiveNameExpression = '$this->name';
		}
		
		eval($this->_Hook_Constructor_post);
	}

	/**
	 * Retrieve row from the database
	 * 
	 * This function retrieves a row from the database, according to the ID of 
	 * the primary key held in the object. The member variables of the object
	 * are updated according to the columns in the database.
	 * 
	 * @param string $key 
	 * @return bool Returns true if a row was successfully retrieved, false otherwise
	 * @access public
	 */
	function getDetails ($Key = DBROW_KEY_NONE)
	{
		global $PHPOF_INITIALISE_PROPERTIES_BEFORE_FETCH, $PHPOF_INITIALISE_PROPERTIES_TO_NULL;
		
		if (!$this->_table->_key) trigger_error('DBRow::getDetails() called but table has no primary key',E_USER_ERROR);
		
		eval($this->_Hook_GetDetails_pre);
		
		$this->RetrievedOK = false;
		
		if ($PHPOF_INITIALISE_PROPERTIES_BEFORE_FETCH) {
			$Type = DBROW_UNSET;
			if ($PHPOF_INITIALISE_PROPERTIES_TO_NULL) $Type = DBROW_SET_TO_NULL;
			$this->clearFields($Type);
		}
		
		if ($Key != DBROW_KEY_NONE) {
			$this->{$this->_table->_key} = $Key;
		}
		
		$return_val = false;

		// First deal with NON-binary fields
		$sql = 'SELECT '.$this->_table->_fieldlist.
			' FROM '.$this->_table->_name.
			' WHERE '.$this->_table->_key ."='".$this->_table->_db->qstr($this->{$this->_table->_key})."'";
		$res = $this->_table->_db->query ($sql);
		if ($res->numRows() > 0) {
			$resultarray = $res->fetchRow($res);
			foreach ($resultarray as $resultkey => $resultval) {
				// We used to get 2x the right number of results here; for each col we got num->value and name->value
				// I think this is fixed now, in DB_xxxx. TODO: Test with postgres.
				// print "resultkey:$resultkey, resultval:$resultval";
				$this->$resultkey = $resultval;
				if ($this->_table->AuditLogging) {
					$this->_audit[$resultkey] = $resultval;
				}
			}
			$return_val = true;
			$this->RetrievedOK = true;
		}
		
		// Now deal with binary fields
		$this->resolveBinaryFields();
		
		$this->setDescriptiveName();
		
		eval($this->_Hook_GetDetails_post);
		return $return_val;
	}
	
	function resolveBinaryFields()
	{
		if (!empty($this->_table->_binaryfields)) {
			foreach ($this->_table->_binaryfields as $field) {
				$sql = 'SELECT COUNT(*) AS filled'.
						' FROM '.$this->_table->_name;
				if ($this->_table->_db->_dbtype == 'mssql') { // HACK!!!
					$sql .= " WHERE $field IS NOT NULL";
				} else {
					$sql .= " WHERE $field != ''";
				}
				$sql .= ' AND '.$this->_table->_key ."='" . $this->_table->_db->qstr($this->{$this->_table->_key}) . "'";
				$res = $this->_table->_db->query($sql);
				$resultarray = $res->fetchRow($res);
				if ($resultarray['filled']) {
					$this->$field = PHPOF_BINARYDATA_PRESENT;
				} else {
					$this->$field = null;
				}
			}
		}
	}
	
	/**
	 * Clears all object fields which relate to columns in the database. Sets them to NULL if $Type is DBROW_SET_TO_NULL, otherwise unsets them.
	 * @param const $Type The type of clear to perform. Defaults to DBROW_UNSET, but may also be DBROW_SET_TO_NULL
	 * @param const $KeyAction  Whether to clear the primary key or not. Should be one of DBROW_KEY_CLEAR or DBROW_KEY_SAVE (the default)
	 * @return void
	 * 
	 */
	function clearFields ($Type=null, $KeyAction=DBROW_KEY_SAVE)
	{
		global $PHPOF_INITIALISE_PROPERTIES_TO_NULL;
		
		if ($Type == null) {
			if ($PHPOF_INITIALISE_PROPERTIES_TO_NULL) {
				$Type = DBROW_SET_TO_NULL;
			} else {
				$Type = DBROW_UNSET;
			}
		}
		
		// Clear key, if appropriate
		if ($KeyAction == DBROW_KEY_CLEAR) {
			if (!empty($this->_table->_key)) {
				if ($Type == DBROW_SET_TO_NULL) {
					$this->{$this->_table->_key} = null;
				} else {
					unset($this->{$this->_table->_key});
				}
			}
		}
		
		if (count($this->_table->_fields) > 0) {
			foreach ($this->_table->_fields as $field) {
				if ($Type == DBROW_SET_TO_NULL) {
					$this->$field = null;
				} else {
					unset($this->$field);
				}
			}
		}
		
		if (isset($this->_table->_binaryfields) && count($this->_table->_binaryfields) > 0) {
			foreach ($this->_table->_binaryfields as $field) {
				if ($Type == DBROW_SET_TO_NULL) {
					$this->$field = null;
				} else {
					unset($this->$field);
				}
			}
		}
	}
	
	/**
	 * Save the current content of all the current non-binary fields into an internal storage space so it can be used for comparison later when constructing audit logs
	 * 
	 * @access private
	 * @return void
	 */
	function _storeAuditData () {
		if ($this->_table->AuditLogging && !$this->_skip_audit) {
    		if (count($this->_table->_fields) > 0) {
    			foreach ($this->_table->_fields as $field) {
    				$this->_audit[$field] = $this->$field;
    			}
    		}
    	}
	}
	
	/**
	 * Returns the contents of column $column_name directly, according to the current primary key
	 * 
	 * Intended for use with BLOB types.
	 * 
	 * @param string $column_name  The name of the column to fetch data from
	 * @param const $query  The query to be performed. Has two possible values: DBROW_GET (default) - retrieve the data from the database, which will be returned from this function, and  DBROW_CHECKONLY - simply check to see if the column is empty or not. This function will return TRUE or FALSE accordingly.
	 * @return bool|string  The binary data from the database, or a boolean if $query is set to DBROW_CHECKONLY. False on error.
	 * @access public
	 */
	function getBinaryData ($column_name, $query=DBROW_GET)
	{
		$sql = "SELECT $column_name".
				 ' FROM '.$this->_table->_name.
				 ' WHERE '.$this->_table->_key ."='".$this->_table->_db->qstr($this->{$this->_table->_key})."'";
		$res = $this->_table->_db->query ($sql);
		if ($res->numRows() > 0) {
			$resultarray = $res->fetchRow($res);
			if ($query == DBROW_CHECKONLY) { 
				if (empty($resultarray[$column_name])) {
					return false;
				} else {
					return true;
				}
			} else {
				return _DBRow_RetrieveBinaryData($this, $column_name, $resultarray);
			}
		} else {
			return false;
		}
	}
	
	
	/**
	 * Store given binary raw data into a column along with the data type.
	 * 
	 * @param $Data string The binary data itself
	 * @param $DataColumn string The name of the column which will receive the binary data from the uploaded file
	 * @param $TypeColumn string The name of the column which will be filled with the MIME type of the data
	 * @param $MIMEType string The MIME type of the data
	 * @access public
	 * @return bool
	 */
	function storeBinaryData(&$Data, $DataColumn, $TypeColumn='', $MIMEType='')
	{
		// TODO: make $MIMEType optional and auto-find the datatype
		if (empty($this->{$this->_table->_key})) {
			trigger_error('DBRow::storeBinaryData() called on object which has not yet been stored in the database',E_USER_WARNING);
			return false;
		}
		
		_DBRow_updateBinaryDataFromBuffer($this, $DataColumn, $TypeColumn, $Data, $MIMEType, $this->_table->_db->_dbh);
		$this->$DataColumn = PHPOF_BINARYDATA_PRESENT;
		if ($TypeColumn) $this->$TypeColumn = $MIMEType;
	}
	
	/**
	 * Handle data upload from file. Replacement for the older handleDataUpload() function
	 * 
	 * @param string $filename  Filesystem-relative filename of the file to upload
	 * @param string $dest_column_name_data  The name of the column which will receive the binary data from the uploaded file
	 * @param string $dest_column_name_type  The name of the column which will be filled with the MIME type of the uploaded data
	 * @param mixed $acceptable_filetypes  List of MIME types which are acceptable for the uploaded data. Can be a single type, or an array of strings.
	 * @param int $maxsize  The maximum acceptable size of the file, in bytes
	 * @param string $UseFilename  Original filename on client system
	 * @return mixed  True if operation succeeded; otherwise a string containing the error. Also returns true if $filename is empty
	 * @access public
	 */
	function handleUpload ($filename, $dest_column_name_data, $dest_column_name_type, $acceptable_filetypes=PHPOF_BINARYDATA_ANY_TYPE, $maxsize=PHPOF_BINARYDATA_UNLIMITED, $UseFilename=null)
	{
		global $MIME_TYPE_MAPPINGS;	// Do NOT remove this or the functions in MIMETypeHelper will not be able to access it 
		
		if (!empty($filename)) {
			require_once('MIMETypeHelper.class.php');
			$filetype = MIMETypeHelper::getTypeFromFile($filename, $UseFilename);
			
			if ($acceptable_filetypes == PHPOF_BINARYDATA_ANY_TYPE || in_array($filetype,$acceptable_filetypes)) {
				$Filesize = filesize($filename);
				if ($Filesize == 0) {
					return 'Empty file uploaded!';
				}
				
				if ($maxsize == PHPOF_BINARYDATA_UNLIMITED || $Filesize <= $maxsize) {
					_DBRow_UpdateBinaryData($this, $dest_column_name_data, $dest_column_name_type, $filename, $filetype, $this->_table->_db->_dbh);
					$this->$dest_column_name_data = PHPOF_BINARYDATA_PRESENT;
					if ($dest_column_name_type) $this->$dest_column_name_type = $filetype;
				} else {
					return "File is too large. Max size is $maxsize bytes; the file you uploaded was $Filesize bytes.";
				}
			} else {
				return "The file you uploaded was detected as the following type: '$filetype' which is not one of the allowed types for this purpose. Please contact an administrator if this is incorrect";
			}
		}
		return true;
	}
	

	/**
	 * Takes data from a standard PHP upload array and handle it, ultimately storing it in the database if all is well.
	 *
	 * Deprecated in favour of DBRow::handleUpload()
	 * 
	 * @param array $postdata PHP POST data
	 * @param string $dest_column_name_data  The name of the column which will receive the binary data from the uploaded file
	 * @param string $dest_column_name_type  The name of the column which will be filled with the MIME type of the uploaded data
	 * @param mixed $acceptable_filetypes  List of MIME types which are acceptable for the uploaded data. Can be a single type, or an array of strings. Set to "any" (default) if you don't care.
	 * @param int $maxsize  The maximum acceptable size of the file, in bytes
	 * @param int $max_x  The maximum acceptable horizontal size of the image, if the data is an image.
	 * @param int $max_y  The maximum acceptable vertical size of the image, if the data is an image.
	 * @return bool|string  TRUE if OK; a string containing the error otherwise
	 */
	function handleDataUpload ($postdata, $dest_column_name_data, $dest_column_name_type, $acceptable_filetypes='any', $maxsize=0, $max_x=0, $max_y=0)
	{
		if ($postdata['tmp_name'] && $postdata['tmp_name'] != 'none') {
			if (filesize($postdata['tmp_name']) == 0) {
				return 'Empty file uploaded!';
			}
			// FIXME: 'file' is a bit shite and doesn't always return MIME types (e.g. PDF)
			$filetype = trim(`file -ib $postdata[tmp_name]`);
			// Fix for 'file' being crap
			if ($filetype == 'image/x-png') $filetype = 'image/png';
			if (in_array($filetype,$acceptable_filetypes) || $acceptable_filetypes == 'any') {
				if ($postdata['size'] <= $maxsize || $maxsize == 0) {
					// Check dimensions
					if ($max_x || $max_y) {
						$dimensions = getimagesize($postdata['tmp_name']);
						if ($dimensions[0] > $max_x || $dimensions[1] > $max_y) {
							return "Incorrect image dimensions. Uploaded image was $dimensions[0] x $dimensions[1]; max is $max_x (width) x $max_y (height)";
						}
					}
					_DBRow_UpdateBinaryData($this, $dest_column_name_data, $dest_column_name_type, $postdata['tmp_name'], $filetype, $this->_table->_db->_dbh);
					$this->$dest_column_name_data = PHPOF_BINARYDATA_PRESENT;
					$this->$dest_column_name_type = $filetype;
				} else {
					return "File is too large. Max size is $maxsize bytes; the file you uploaded was ".$postdata['size']." bytes.";
				}
			} else {
				return "Wrong filetype (detected as '$filetype'; contact an administrator if this is wrong)";
			}
		}
	}
	
	/**
	 * Fills the properties of current object from an associative array, where the array contains "column name" => "column content" pairs
	 * 
	 * This function takes an associative array array and sets the DBRow object
	 * parameters from it. For example, if array was an associative array like 
	 * so:
	 * 
	 * $myarray = array ('foo'=>'bar', 'fish'=>'haddock');
	 * 
	 * then calling DBRow::fillFromArray($myarray) would result in the DBRow 
	 * object parameter 'foo' being filled with the string 'bar', and the
	 * parameter 'fish' being filled with the string 'haddock'.
	 * 
	 * This can be particularly useful when taking data from an HTML form. For 
	 * example, if you had a user-submitted form with a set of parameters (first
	 * name, last name, age, sex etc.) and these corresponded with the columns 
	 * in a database table, you could write some HTML like this:
	 * 
	 * <input name="param[first_name]">
	 * <input name="param[last_name]">
	 * <input name="param[age]">
	 * 
	 * and then capture this with a PHP form and fill an object with it like this:
	 * 
	 * $MyDBRowObject->fillFromArray($param);
	 * 
	 * @param array $array  Data to be stored
	 * @return void
	 */
	function fillFromArray ($array)
	{
		if (is_array($array)) {
			if (count($array) >0 ) {
				foreach ($array as $key => $value) {
					$this->$key = $value;
				}
			}
		} else {
			trigger_error('DBRow::fillFromArray was passed a non-array',E_USER_WARNING);
		}
	}
	
	/**
	 * Stores a row in the database. Inserts if a row with the matching primary key already exists, otherwise updates existing row.
	 * 
	 * It is suitable for situations where you do not know whether the object 
	 * you are dealing with already exists as a row or whether it is a new row 
	 * which must be inserted. If you already know whether you want to update or
	 * insert, you would be better off using DBRow::insert() or DBRow::update()
	 * 
	 * The function tries to act intelligently and choose the appropriate
	 * action, based on the following factors:
	 * - Whether the table which the row belongs to has a primary key
	 * - Whether a value is supplied for the primary key
	 * - Whether the primary key field in the table (where applicable) is auto-incrementing
	 * 
	 * @return void
	 */
	function store ()
	{
		eval($this->_Hook_Store_pre);
		if ($this->_table->_key) {								// If the table has a primary key column
			if ($this->_table->_key_auto_increment) {		// ...then if the table has an auto-incrementing key column
				if (!empty($this->{$this->_table->_key})) {			// ......then if the key is supplied
					$this->Update();								// .........Update an existing row according to the key
				} else {												// ......Otherwise, 
					$this->Insert();								// .........Insert a new row (creating a new key id)
				}
			} else {													// There is no auto-inc key column,
				if ($this->{$this->_table->_key}) {			// so check if a key is supplied.
					if ($this->_RowExists()) {					// ...If it is, see if a row exists with that key.
						$this->Update();							// ......If it does, update that row.
					} else {											// ...Otherwise,
						$this->Insert();							// ......Insert a new row with the given key
					}
				} else {												// There's no auto-inc key column and no key was supplied
					trigger_error('DBRow::Store() called but no key supplied and key column is not auto-incrementing',E_USER_ERROR); // don't know what to do
				}
			}
		} else {
			trigger_error('DBRow::Store() called on a table with no key column; data will be inserted but you should use DBRow::insert() instead to avoid ambiguity',E_USER_NOTICE);
			$this->Insert();
		}	

		eval($this->_Hook_Store_post);
	}
	
	/**
	 * Find out if a row exists in the database with a key matching the current one
	 */
	function _rowExists ()
	{
		// TODO: could we replace this with $res = $this->_table->BasicSearch($this->_table->_key,$this->{$this->_table->_key}); ?
		$res = $this->_table->_db->query("SELECT * FROM ".$this->_table->_name.' WHERE '.$this->_table->_key."='".$this->_table->_db->qstr($this->{$this->_table->_key})."'");
		if ($res->numRows() >0 ) {
			return true;
		} else {
			return false;
		}
	}
	
	/**
	 * Insert a row into the database
	 *
	 */
	function insert ()
	{
		eval($this->_Hook_Insert_pre);
		$SQL = 	'INSERT INTO '.$this->_table->_name. '(';
		
		// If the table has a key, but it is not auto-incrementing, add the key column name to the list of fields to be inserted
		if ($this->_table->_key && !$this->_table->_key_auto_increment) $SQL .= $this->_table->_key.",";
		
		// Add a sort order if this table has a sort field and the current item doesn't have one
		if (!empty($this->_table->_SortOrder) && empty($this->{$this->_table->_SortOrder})) {
			if (empty($this->_table->_SortDefaultPosition)) {
				$this->_table->_SortDefaultPosition = DBROW_SORT_BOTTOM;
			}
			$this->{$this->_table->_SortOrder} = $this->_table->_SortDefaultPosition;
		}
		
		// List the fields which are to be inserted
		// TODO: Surround the fieldnames in double quotes if we're using Postgres, to catch the case of reserved words as column names...dunno about MySQL
		$SQL .=	$this->_table->_fieldlist;
				
		$SQL.=') VALUES (';
		
		// If the table has a key, add the key ID (only if not auto-incrementing)
		if ($this->_table->_key && !$this->_table->_key_auto_increment) $SQL .= "'" . $this->_table->_db->qstr($this->{$this->_table->_key}) . "',";
		
		// Add the values of the fields to be inserted to the SQL query
		if (count($this->_table->_fields) > 0) {
			foreach ($this->_table->_fields as $field) {
				$value = "'" . $this->_table->_db->qstr($this->$field) . "'";
				if ($value=="''") $value = 'NULL';
				$SQL .= "$value,";
			}
			$SQL = substr($SQL,0,-1);	// strip the trailing comma
		}
		
		$SQL .=	')';

		$res = $this->_table->_db->query($SQL);
		
		if ($this->_table->_key_auto_increment) {
			$this->{$this->_table->_key} = $res->insertedId($this->_table->_name);
		}
		
		if ($this->_table->AuditLogging && !$this->_skip_audit) $this->_auditLog(PHPOF_EVENT_INSERT); // FIXME?: only if query succeeded
		
		if (!empty($this->_table->_SortOrder)) $this->processSortOrder();
		
		eval($this->_Hook_Insert_post);
	}
	
	/**
	 * Update an existing row in database, based on the currently-set primary key
	 * 
	 * @return void
	 */
	function update ()
	{
		eval($this->_Hook_Update_pre);
		if (!empty($this->{$this->_table->_key})) {
			// Handle binary fields
			if (!empty($this->_table->_binaryfields)) {
				foreach ($this->_table->_binaryfields as $field) {
					if ($this->$field == PHPOF_BINARYDATA_DELETE) {
						_DBRow_RemoveBinaryData($this, $field, $this->_table->_db->_dbh);
						$this->$field = '';
					}
				}
			}
			
			// Create the SQL query
			// TODO: only update changed cols
			$SQL = 'UPDATE '.$this->_table->_name.' SET ';
			foreach ($this->_table->_fields as $field) {
				$value = "'" . $this->_table->_db->qstr($this->$field) . "'";
				if ($value=="''") $value = 'NULL';
				if ($this->_table->_db->_dbtype == 'pgsql') {
					$SQL .= "\"$field\"" . "=$value,"; // Need to quote column names in Postgres in case they are a reserved word
				} else {
					$SQL .= $field . "=$value,"; // Not sure what happens in MySQL if you use a reserved word as a column name
				}
			}
			$SQL = substr($SQL,0,-1);	// strip the trailing comma
			$SQL .=	' WHERE '.$this->_table->_key." = '" . $this->_table->_db->qstr($this->{$this->_table->_key}) . "'";
		
			$this->_table->_db->query($SQL);			
		} else {
			if ($this->_table->_key) {
				trigger_error("DBRow::Update() called, but the primary key field ($this->_table->_key) is empty",E_USER_ERROR);
			} else {
				trigger_error("DBRow::Update() called, but table ($this->_table->_name) does not have a primary key field",E_USER_ERROR);
			}
		}
		
		if ($this->_table->AuditLogging && !$this->_skip_audit) $this->_auditLog(PHPOF_EVENT_UPDATE);	// FIXME?: only if query succeeded
		
		if (!empty($this->_table->_SortOrder)) $this->processSortOrder();
		
		eval($this->_Hook_Update_post);
	}
	
	/**
	 * Delete row from database
	 * 
	 * @return void
	 */
	function delete ()
	{
		eval($this->_Hook_Delete_pre);
		
		if (!empty($this->_table->_key)) {
			if (!empty($this->{$this->_table->_key})) {
				// Remove any binary data from columns in the row
				if (!empty($this->_table->_binaryfields)) {
					foreach ($this->_table->_binaryfields as $field) {
						_DBRow_RemoveBinaryData($this, $field, $this->_table->_db->_dbh);
					}
				}
				
				// Remove the actual row from the database
				// TODO?: Need add/stripslashes for $this->{$this->_table->_key} ?
				// TODO?: Should probably quote $this->_table->_key for Postgres...
				$this->_table->_db->query('DELETE FROM '.$this->_table->_name.
											' WHERE '.$this->_table->_key ."='".$this->_table->_db->qstr($this->{$this->_table->_key})."'");
			} else {
				trigger_error('DBRow::delete() called but row key ('.$this->_table->_key.') is empty',E_USER_ERROR);
			}
		} else {
			// No table key - delete rows which match ALL columns
			$sql = 'DELETE FROM '.$this->_table->_name.' WHERE ';
			foreach ($this->_table->_fields as $field) {
				if ($this->$field === null) {
					$sql .= "$field IS NULL AND ";
				} else {
					$sql .= "$field='" . $this->_table->_db->qstr($this->$field) . "' AND ";
				}
			}
			$sql = substr($sql,0,-5);	// strip trailing " AND "
			$this->_table->_db->query($sql);
		}
		
		$this->ClearFields();
		if ($this->_table->AuditLogging && !$this->_skip_audit) $this->_auditLog(PHPOF_EVENT_DELETE);	//FIXME?: only if query succeeded

		eval($this->_Hook_Delete_post);
	}
	
	/**
	 * This function sets an 'owner' identifier for the next action for which 
	 * audit logging takes place (typically an update, insert or delete). 
	 * The identifier can take any form you choose; it is stored verbatim in the
	 * audit_events table.
	 * 
	 * This is only applicable if you have audit logging enabled (config option PHPOF_AUDIT_LOGGING).
	 *
	 * @param string $owner
	 */
	function setAuditOwner ($owner)
	{
		$this->_audit_owner = $owner;
	}
	
	/**
	 * This function sets the descriptive notes which will be stored for the 
	 * next action for which audit logging takes place (typically an update,
	 * insert or delete). The notes can take any form you choose; they are 
	 * stored verbatim in the audit_events table.
	 * 
	 * This is only applicable if you have audit logging enabled (config option 
	 * PHPOF_AUDIT_LOGGING).
	 *
	 * @param string $notes
	 */
	function setAuditNotes ($notes)
	{
		$this->_audit_notes = $notes;
	}
	
	/**
	 * Add a (standalone) audit note to the database, not linked to any specific changes
	 * 
	 * @param string $note  The note to add
	 * @param string $owner  "Owner". PHPOF does not interpret this in any way.
	 * @return void
	 */
	function addAuditNote($note, $owner=null)
	{
		$audit_events = new AuditEvents();
		$audit_event = new AuditEvent($audit_events);
		
		$audit_event->owner = $owner;
		$audit_event->table_name = $this->_table->_name;
		$audit_event->event_type = PHPOF_EVENT_NOTE;
		$audit_event->key_value = $this->{$this->_table->_key}; // TODO: what if no key?
		$audit_event->timestamp = date('Y-m-d H:i:s');	// TODO?: change col name to 'date_time' instead of 'timestamp'?
		$audit_event->notes = $note;
		$audit_event->insert();

		return true;
	}
	
	function _auditLog ($event_type)
	{
		global $PHPOF_HOOK_AUDIT_PRE;
 		eval($PHPOF_HOOK_AUDIT_PRE);
		$audit_actions = new AuditActions();
		$audit_action = new AuditAction($audit_actions);
		
		// Check which (if any) columns have actually been changed, and
		// build the appropriate AuditAction objects
		$actions = 0;
		foreach ($this->_table->_fields as $field) {
			if (!in_array($field, $this->_table->AuditColumnsIgnore)) {
				if ($this->$field != $this->_audit[$field]) {
					$audit_action->column_name = $field;
					$audit_action->prev_value = stripslashes($this->_audit[$field]);
					$audit_action->new_value = stripslashes($this->$field);	// shouldn't really need stripslashes() here; if the col has been edited then it shouldn't have slashes in. Better to be safe than sorry though.
					if (PHP_VERSION > '5.0.0') {
						// Need to clone the object. Use eval() as "clone" doesn't parse in PHP4
						eval('$audit_actions_to_store[] = clone $audit_action;');
					} else {
						$audit_actions_to_store[] = $audit_action;
					}
					$actions++;
				}
			}
		}
		
		if ($actions > 0 || $event_type == PHPOF_EVENT_NOTE) {
			$this->_table->_db->beginTransaction();
			
			// Set up audit event
			$audit_events = new AuditEvents();
			$audit_event = new AuditEvent($audit_events);
		
			$audit_event->owner = $this->_audit_owner;
			$audit_event->table_name = $this->_table->_name;
			$audit_event->event_type = $event_type;
			$audit_event->key_value = $this->{$this->_table->_key}; // TODO: what if no key?
			$audit_event->timestamp = date('Y-m-d H:i:s');	// TODO?: change col name to 'date_time' instead of 'timestamp'?
			$audit_event->notes = $this->_audit_notes;
			$audit_event->Insert();
			$this->_audit_notes = '';	// To ensure future queries don't accidentally get these notes.
			
			if ($actions > 0) {
				foreach ($audit_actions_to_store as $audit_action) {
					$audit_action->audit_event_id = $audit_event->id;
					$audit_action->Insert();
				}
			}
			$this->_table->_db->commitTransaction();
		}
	}

	/* ---------------------------------------------------------------------------------
		SORT ORDER STUFF
	--------------------------------------------------------------------------------- */
	
	function move($Direction, $Num=1)
	{
		if (empty($this->_table->_SortOrder)) trigger_error('DBRow::move() called but object\'s table does not have a sort column defined',E_USER_ERROR);
		
		if ($Num > 1) {
			$this->move($Direction, ($Num-1));
		}
		
		// This is an additional qualifier for if a table has got several different sequences of sort
		// orders within it
		$Qualifier = $this->_getSortOrderQualifierSQL();
		
		switch ($Direction) {
			case DBROW_MOVE_UP:
				$WhereClause = '<';
				$OrderDirection = 'DESC';
			break;
			
			case DBROW_MOVE_DOWN:
				$WhereClause = '>';
				$OrderDirection = 'ASC';
			break;
			
			default:
			trigger_error("Invalid direction in DBRow::move() [Supplied: $Direction]",E_USER_ERROR);
		}
		
		// First see if the ordering is fully expanded (i.e. all items have a unique sort order value)
		$TotalItems = $this->_table->_db->query("SELECT COUNT(*) AS count FROM ".$this->_table->_name." WHERE 1 $Qualifier");
		$TotalItems = $TotalItems->fetchRow();
		$TotalItems = $TotalItems['count'];
		
		$UniqueItems = $this->_table->_db->query("SELECT COUNT(DISTINCT ".$this->_table->_SortOrder.") AS count FROM ".$this->_table->_name." WHERE 1 $Qualifier");
		$UniqueItems = $UniqueItems->fetchRow();
		$UniqueItems = $UniqueItems['count'];
		
		if ($TotalItems != $UniqueItems) {
			if (empty($this->_table->_key)) trigger_error('DBRow::sort() called, but the table (with qualifiers) does not contain a set of unique sort values and the table does not have a primary key column. The behaviour of this function is undefined.',E_USER_WARNING);
			$this->_sort($Qualifier);
		}
		
		$AdjacentItem = $this->_table->_db->query("SELECT * FROM ".$this->_table->_name." WHERE ".$this->_table->_SortOrder." $WhereClause ".$this->{$this->_table->_SortOrder}." $Qualifier ORDER BY ".$this->_table->_SortOrder." $OrderDirection LIMIT 1");
		if ($AdjacentItem->numRows() > 0) {
			$Item = new DBRow($this->_table);
			
			// Move the adjacent item to the current item's position
			$AdjacentItem->fetchIntoObject($Item);
			$AdjacentOrder = $Item->{$this->_table->_SortOrder};
			$Item->{$this->_table->_SortOrder} = $this->{$this->_table->_SortOrder};
			$Item->update();
			
			// Move current item to its final position
			$this->{$this->_table->_SortOrder} = $AdjacentOrder;
			$this->update();
		} else {
			// Already at the top
			return false;
		}
	}
	
	function _getSortOrderQualifierSQL()
	{
		$Qualifier = '';
		if (!empty($this->_table->_SortOrderQualifierColumns) && is_array($this->_table->_SortOrderQualifierColumns)) {
			foreach ($this->_table->_SortOrderQualifierColumns as $ColumnName) {
				$Qualifier = "AND $ColumnName=".$this->_table->_db->qstr($this->$ColumnName);
			}
		}
		return $Qualifier;
	}
	
	function processSortOrder()
	{
		$SortField = $this->_table->_SortOrder;
		if (!empty($this->$SortField) && !empty($this->_table->_key)) {
			if ($this->$SortField == DBROW_SORT_TOP || $this->$SortField == DBROW_SORT_BOTTOM) {

				$Qualifier = $this->_getSortOrderQualifierSQL();

				if ($this->$SortField == DBROW_SORT_TOP) {
					// Sort to top by setting a priority ordering factor of "<key>=<id> DESC"
					$this->_sort($Qualifier, $this->_table->_key.'='.$this->id.' DESC');
				} else {
					// Sort to bottom
					$Res = $this->_table->_db->query("SELECT MAX($SortField) AS max FROM ".$this->_table->_name." WHERE 1 $Qualifier");
					$Row = $Res->fetchRow();
					$this->$SortField = $Row['max']+1;
					$this->update();
				}
			}
		}
	}
	
	function _sort($Qualifier, $PriorityOrder='')
	{
		if (!empty($PriorityOrder)) $PriorityOrder.=',';
		$AllItems = $this->_table->_db->query('SELECT * FROM '.$this->_table->_name." WHERE 1 $Qualifier ORDER BY $PriorityOrder ".$this->_table->_SortOrder.' ASC');

		if ($AllItems->numRows() > 0) {
			$Item = new DBRow($this->_table);
			$OrderValue = 1;
			while ($AllItems->fetchIntoObject($Item)) {
				if ($Item->{$this->_table->_key} == $this->{$this->_table->_key}) {
					$NewOrderValue = $OrderValue;
				}
				$Item->{$this->_table->_SortOrder} = $OrderValue++;
				$Item->update();
			}
			
			// Update this object with it's new sort order value
			$this->{$this->_table->_SortOrder} = $NewOrderValue;
		}
	}
	
	/* ---------------------------------------------------------------------------------
		DESCRIPTIVE NAME STUFF
	--------------------------------------------------------------------------------- */
	function getDescriptiveName ()
	{
		if (!empty($this->DescriptiveNameExpression)) {
			eval("\$DescriptiveName = \"$this->DescriptiveNameExpression\";");
			$DescriptiveName = str_replace('()','',$DescriptiveName);
			$DescriptiveName = str_replace('[]','',$DescriptiveName);
			return trim($DescriptiveName);
		}
		return null;
	}
	
	function setDescriptiveName()
	{
		if (!empty($this->DescriptiveNameExpression)) {
			$this->DescriptiveName = $this->getDescriptiveName();
			return $this->DescriptiveName;
		}
		return null;
	}

}
