<?php

// DB_pgsql.class.php - abstract database functions and make them easier to use
// Attempts to broadly retain syntactic compatibility with PEAR DB class.
// PostgreSQL version
//
// Version $Id: DB_pgsql.class.php 2258 2017-09-07 20:50:08Z tim $
//
// (c)2001-2017 Tim Jackson (tim@timj.co.uk)
//
// This library is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public
// License as published by the Free Software Foundation; either
// version 2.1 of the License, or (at your option) any later version.
//
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// Lesser General Public License for more details.
// 
// You should have received a copy of the GNU Lesser General Public
// License along with this library; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

require_once('pgsql.php');

/**
 * PostgreSQL driver for PHPOF
 * See the MySQL driver for code documentation
 *
 */
class DB_pgsql {

	var $_dbh;					// Database handle
	var $_dbtype='pgsql';	// Database type
	var $_dbhost;				// Database host
	var $_dbname;				// Database name

	// Connect to the database server and select the database we wish to use
	function DB_pgsql ($dbhost, $username, $password, $dbname, $charset=null) {
		$this->_dbh = @pg_connect("host=$dbhost dbname=$dbname user=$username password=$password") or trigger_error('Could not connect to database',E_USER_ERROR);
		$this->_dbname = $dbname;
		$this->_dbhost = $dbhost;
		if ($charset === null) $charset = 'LATIN1';
		pg_set_client_encoding($this->_dbh, $charset);
	}

	// Query the database
	function query ($query) {
		global $PHPOF_DEBUG_MODE;
		if ($PHPOF_DEBUG_MODE) phpof_debug("DB_pgsql::query(): $query");
		$result = @pg_query ($this->_dbh, $query);
		
		if ($err = pg_last_error()) {
			if (strlen($query) > 2048) { // truncate for sanity, otherwise can get huge binaries in error reports
				$query = substr($query, 0, 2048) . ' [truncated]';
			}
			$err .= " (SQL query was: $query)";
			trigger_error($err,E_USER_ERROR);
		}
		
		$resobj = new DB_result($this, $result);	// *NOT* $this->_dbh for Postgres!!
		return $resobj;
	}
	
	// Perform a query, returning results from a particular page
	function query_paged ($query, $results_per_page, $page) {
		global $PHPOF_DEBUG_MODE;
		if ($page < 1) trigger_error("DB::query_paged passed a page number ($page) less than 1", E_USER_ERROR);
		$base = $results_per_page * ($page - 1);
		$fullquery = "$query LIMIT $results_per_page, $base";
		
		if ($PHPOF_DEBUG_MODE) phpof_debug("DB_pgsql::query_paged(): $fullquery");
		
		$result = @pg_query ($this->_dbh,$fullquery);
		
		if ($err = pg_last_error()) {
			if (strlen($fullquery) > 2048) { // truncate for sanity, otherwise can get huge binaries in error reports
				$query = substr($fullquery, 0, 2048) . ' [truncated]';
			}
			$err .= " (SQL query was: $fullquery)";
			trigger_error($err,E_USER_ERROR);
		}
		
		$resobj = new DB_result($this, $result);	// *NOT* $this->_dbh for Postgres!!
		$resobj->pages = $this->query_num_pages($query, $results_per_page);
		
		return $resobj;
	}
	
	// Get the number of pages that would be returned for a given query
	function query_num_pages ($query, $results_per_page) {
		global $PHPOF_DEBUG_MODE;
		if ($results_per_page == 0) trigger_error('DB_pgsql::query_num_pages() called with zero results per page',E_USER_ERROR);
	
		if ($PHPOF_DEBUG_MODE) phpof_debug("DB_pgsql::query_num_pages(): $query");
		$result = @pg_query ($this->_dbh, $query);
		
		if ($err = pg_last_error()) {
			if (strlen($query) > 2048) { // truncate for sanity, otherwise can get huge binaries in error reports
				$query = substr($query, 0, 2048) . ' [truncated]';
			}
			$err .= " (SQL query was: $query)";
			trigger_error($err,E_USER_ERROR);
		}
		
		$resobj = new DB_result($this, $result);	// *NOT* $this->_dbh for Postgres!!
		$rows = $resobj->numRows();
		
		return ceil ($rows / $results_per_page);
	}
	
	function beginTransaction()
	{
		$this->query('BEGIN');
	}
	
	function rollbackTransaction()
	{
		$this->query('ROLLBACK');

	}
	
	function commitTransaction()
	{
		$this->query('COMMIT');
	}
}

class DB_result {

	var $_dbh;				// Database handle
	var $_db;				// Database object; only for Postgres
	var $_rr;				// Result resource
	var $_currentrow=0;	// Only for Postgres; keep track of current row
	var $pages=1;			// Total pages for this query

	// Constructor
	function DB_result (&$db, $rr) {	// NB this is passed a DB *object*, NOT a DB *handle* as with DB_result for MySQL
		$this->_dbh = &$db->_dbh;
		$this->_rr = $rr;
		$this->_db = &$db;			// Only for Postgres; for use with insertedId()
	}

	// Get the number of rows in the result
	function numRows () {
		return pg_num_rows($this->_rr);	// will be pg_num_rows
	}
	
	// Fetch a row from the result set as an associative array
	function fetchRow () {
		// Need '@' in the below line, or it will return a warning after the last row of every result set
		return @pg_fetch_array($this->_rr, $this->_currentrow++, PGSQL_ASSOC);	
	}

	// $table_name is an optimisation introduced in V1.0.7 which dramatically improves
	// the speed when called from DBRow::Insert() by pre-selecting the table
	function insertedId ($table_name='') {
		global $PHPOF_PGSQL_DISABLE_INSERTEDID;
		if ($PHPOF_PGSQL_DISABLE_INSERTEDID == true) return PHPOF_DISABLED;
		$lastoid = pg_last_oid($this->_rr);
		if ($table_name) {
			return $this->_insertedId_findIdFromOid($table_name,$lastoid);
		} else {
			// Find which table $lastoid is in by brute force
			// There really ought to be a better way of doing this, but I can't find one.
			$tables = $this->_db->query (	"SELECT c.relname ".				// relname === table name
													"FROM pg_class c ".
													"WHERE c.relkind = 'r'".
													"AND c.relname !~ '^pg_'");	// Find all tables in the database, except internal (pg_*) tables
			if ($tables->numRows() > 0) {
				while ($table = $tables->fetchRow()) {
					$table_name = $table['relname'];
					
					// We could create a DBTable object and then do the below query, but that would involve even more overhead
					$rows = $this->_db->query("SELECT oid FROM $table_name WHERE oid='$lastoid'");	// See if the last OID is in this table
					if ($rows->numRows() > 0) {
						// We have found the table that contains the row with $lastoid.
						// Now find the ID contained in the key column
						return $this->_insertedId_findIdFromOid($table_name, $lastoid);
					}
				}
			}
		}
		return 0;
	}
	
	function _insertedId_findIdFromOid ($table_name, $oid) {
		$table = new DBTable($this->_db,$table_name);
		if ($table->_key) {	
			$row = $table->BasicSearch('oid',$oid);
			$row = $row->fetchRow();
			return $row[$table->_key];
		} else {
			return 0;
		}
	}
	
	function getArray ($key, $value=PHPOF_ARRAY_LINEAR) {
		while ($result = $this->fetchRow()) {
			if ($value == PHPOF_ARRAY_LINEAR) {
				$array[] = $result[$key];
			} else {
				$array[$result[$key]] = $result[$value];
			}
		}
		if (isset($array)) {
			return $array;
		}
		return false;
	}

	function fetchIntoObject (&$object_to_fill) {
		global $PHPOF_AUDIT_LOGGING;
		if ($object_to_fill == '') trigger_error('DB_result::fetchIntoObject() was not passed an object',E_USER_ERROR);
		if ($object_to_fill->_Hook_FetchIntoObject_pre) eval('$object_to_fill->'.$object_to_fill->_Hook_FetchIntoObject_pre);
		if ($row = $this->fetchRow()) {
			foreach ($row as $field => $value) {
				$object_to_fill->$field = $value;
			}
			if ($PHPOF_AUDIT_LOGGING) $object_to_fill->_StoreAuditData();
			
			// Set row object descriptive name
			$object_to_fill->setDescriptiveName();
			
			// Call post-hook
			if ($object_to_fill->_Hook_FetchIntoObject_post) eval('$object_to_fill->'.$object_to_fill->_Hook_FetchIntoObject_post);
			return true;
		} else {
			return false;
		}
	}

	function affectedRows () {
		return pg_affected_rows();
	}
	
}

